import json
import mimetypes

import magic
import mimeparse

from jinja2 import Template
from jinja2.exceptions import TemplateNotFound
from giotto.exceptions import NoViewMethod
from giotto.utils import Mock, htmlize, htmlize_list, pre_process_json

def renders(*mimetypes):
    def decorator(func):
        func.render_map = {}
        for mimetype in mimetypes:
            func.render_map[mimetype] = func.__name__
        return func
    return decorator

class GiottoView(object):
    """
    Base class for all Giotto view objects. All Giotto Views must at least descend
    from this class, as ths class contains piping that the controller calls.
    """

    render_map = {}

    def __init__(self):
        for method in [x for x in dir(self) if not x.startswith('__')]:
            # register all render methods based on the render decorator
            attr = getattr(self, method)
            render_map = getattr(attr, 'render_map', None)
            if render_map and type(render_map) == dict:
                self.render_map.update(render_map)

    def render(self, result, mimetype, errors=None):
        status = 200
        available_mimetypes = self.render_map.keys()
        target_mimetype = mimeparse.best_match(available_mimetypes, mimetype)
        self.errors = errors or Mock()
        renderer = self.render_map[target_mimetype]

        if not renderer:
            raise NoViewMethod("%s not supported for this program" % mimetype)

        data = getattr(self, renderer)(result)

        if target_mimetype == '*/*':
            data['mimetype'] = ''

        if not 'mimetype' in data:
            data['mimetype'] = target_mimetype

        data['status'] = status
        return data

class BasicView(GiottoView):
    """
    Basic viewer that contains generic functionality for showing any data.
    """
    @renders('application/json')
    def json(self, result):
        obj = pre_process_json(result)
        j = json.dumps(obj)
        return {'body': j, 'mimetype': 'application/json'}

    @renders('text/html')
    def html(self, result):
        """
        Try to display any object in sensible HTML.
        """
        css = """
        <style>
            table {border-collapse: collapse}
            td, th {border: 1px solid black; padding: 10px}
            th {background: white;}
            tr:nth-child(even) {background: #DDEE00}
            tr:nth-child(odd) {background: #00EEDD}
        </style>
        """
        h1 = htmlize(type(result))

        if result is None:
            return "<!DOCTYPE html><html><body>None</body></html>"

        out = []
        if not hasattr(result, 'iteritems'):
            header = "<tr><th>Value</th></tr>"
            if type(result) is list:
                result = htmlize_list(result)
            else:
                result = htmlize(result)
            out = ["<tr><td>" + result + "</td></tr>"]
        elif hasattr(result, 'lower'):
            out = ["<tr><td>" + result + "</td></tr>"]
        else:
            # object is a dict
            header = "<tr><th>Key</th><th>Value</th></tr>"
            for key, value in result.iteritems():
                v = htmlize(value)
                row = "<tr><td>{0}</td><td>{1}</td></tr>".format(key, v)
                out.append(row)

        out = "\n".join(out)
        return {'body': """<!DOCTYPE html>
        <html>
            <head>{0}</head>
            <body>
                <h1>{1}</h1>
                <table>
                    {2}
                    {3}
                </table>
            </body>
        </html>""".format(css, h1, header, out),
        'mimetype': 'text/html'}

    @renders('text/x-cmd', 'text/x-irc', 'text/plain')
    def plaintext(self, result):
        out = []
        if hasattr(result, 'iteritems'):
            to_iterate = result.iteritems()
        elif hasattr(result, 'lower'):
            return {'body': result, 'mimetype': "text/plain"}
        else:
            to_iterate = result.__dict__.iteritems()

        for key, value in to_iterate:
            row = "{0} - {1}".format(key, value)
            out.append(row)

        return {'body': "\n".join(out), 'mimetype': "text/plain"}

def JinjaTemplateView(template_name, name='data', mimetype="text/html"):
    class JinjaTemplateView(BasicView):
        @renders(mimetype)
        def html(self, result):
            from giotto import config
            template = config.jinja2_env.get_template(template_name)
            context = {name: result or Mock(), 'errors': self.errors}
            rendered = template.render(**context)
            return {'body': rendered, 'mimetype': mimetype}

    return JinjaTemplateView

class ImageViewer(GiottoView):
    """
    For viewing images. The 'result' must be a file object that contains image
    data (doesn't matter the format).
    """

    @renders('text/plain')
    def plaintext(self, result):
        """
        Converts the image object into an ascii representation. Code taken from
        http://a-eter.blogspot.com/2010/04/image-to-ascii-art-in-python.html
        """
        from PIL import Image

        ascii_chars = ['#', 'A', '@', '%', 'S', '+', '<', '*', ':', ',', '.']

        def image_to_ascii(image):
            image_as_ascii = []
            all_pixels = list(image.getdata())
            for pixel_value in all_pixels:
                index = pixel_value / 25 # 0 - 10
                image_as_ascii.append(ascii_chars[index])
            return image_as_ascii   

        img = Image.open(result)
        width, heigth = img.size
        new_width = 80 
        new_heigth = int((heigth * new_width) / width)
        new_image = img.resize((new_width, new_heigth))
        new_image = new_image.convert("L") # convert to grayscale

        # now that we have a grayscale image with some fixed width we have to convert every pixel
        # to the appropriate ascii character from "ascii_chars"
        img_as_ascii = image_to_ascii(new_image)
        img_as_ascii = ''.join(ch for ch in img_as_ascii)
        out = []
        for c in range(0, len(img_as_ascii), new_width):
            out.append(img_as_ascii[c:c+new_width])

        return "\n".join(out)

    def image(self, result):
        return result

    def image_jpeg(self, result):
        return self.image(result)

    def text_cmd(self, result):
        return result.read()

    def text_html(self, result):
        return """<!DOCTYPE html>
        <html>
            <body>
                <img src="/multiply?x=4&y=7">
            </body>
        </html>"""



class GiottoPrimitive(object): 
    def __init__(self, name):
        self.name = name

LOGGED_IN_USER = GiottoPrimitive("LOGGED_IN_USER")
RAW_PAYLOAD = GiottoPrimitive("RAW_PAYLOAD")
USER_COUNTRY = GiottoPrimitive("USER_COUNTRY")
PREVIOUS_INPUT = GiottoPrimitive("PREVIOUS_INPUT")
PREVIOUS_ERRORS = GiottoPrimitive("PREVIOUS_ERRORS")
ALL_PROGRAMS = GiottoPrimitive('ALL_PROGRAMS')


import inspect
import json

from giotto.programs import GiottoProgram
from giotto.exceptions import InvalidInput, ProgramNotFound, MockNotFound, ControlMiddlewareInterrupt
from giotto.primitives import GiottoPrimitive
from giotto.keyvalue import DummyKeyValue

class GiottoController(object):
    middleware_interrupt = None
    
    def __init__(self, request, manifest, model_mock=False, errors=None):
        from giotto import config
        self.request = request
        self.model_mock = model_mock
        self.cache = config.cache or DummyKeyValue()
        self.errors = errors
        self.manifest = manifest

        # the program that corresponds to this invocation
        invocation = self.get_invocation()
        name = self.get_controller_name()
        parsed = self.manifest.parse_invocation(invocation, name)

        self.program = parsed['program']
        self.program.name = parsed['name']
        self.path_args = parsed['args']
        self.mimetype = parsed['superformat_mime'] or self.mimetype_override() or self.default_mimetype

    def get_response(self):
        control = None
        name = self.get_controller_name()
        try:
            self.request = self.program.execute_input_middleware_stream(self.request, name)
        except ControlMiddlewareInterrupt as exc:
            # A middleware class returned a control object, save it to the class.
            # The get_data_response method will use it.
            self.middleware_interrupt = exc.control

        response = self.get_concrete_response()

        return self.program.execute_output_middleware_stream(self.request, response, name)

    def get_data_response(self):
        """
        Execute the model and view, and handle the cache.
        Returns controller-agnostic response data.
        """
        if self.middleware_interrupt:
            return self.middleware_interrupt

        if self.model_mock:
            model_data = self.program.get_model_mock()
        else:
            args, kwargs = self.program.get_model_args_kwargs()
            data = self.get_data_for_model(args, kwargs)

            if self.program.cache and not self.errors:
                key = self.get_cache_key(data)
                hit = self.cache.get(key)
                if hit:
                    return hit
        
            model_data = self.program.execute_model(data)
        
        response = self.program.execute_view(model_data, self.mimetype, self.errors)

        if self.program.cache and not self.errors and not self.model_mock:
            self.cache.set(key, response, self.program.cache)

        return response

    def get_data_for_model(self, args, kwargs):
        """
        In comes args and kwargs expected for the model. Out comes the data from
        this invocation that will go to the model.
        """
        raw_data = self.get_raw_data()
        output = {}
        for i, arg in enumerate(args):
            if i < len(self.path_args):
                # use path args instead (if they have been supplied)
                target = self.path_args[i]
            else:
                target = raw_data.get(arg, None)

            output[arg] = target

        for key, value in kwargs.iteritems():
            if isinstance(value, GiottoPrimitive):
                output[key] = self.get_primitive(value.name)
            else:
                output[key] = value

        return output


    def __repr__(self):
        controller = self.get_controller_name()
        model = self.get_program_name()
        data = self.get_data()
        return "<%s %s - %s - %s>" % (  
            self.__class__.__name__, controller, model, data
        )

    def mimetype_override(self):
        """
        In some circumstances, the returned mimetype can be changed. Return that here.
        Otherwise the default or superformat will be used.
        """
        return None

    def get_cache_key(self, data):
        try:
            controller_args = json.dumps(data, separators=(',', ':'), sort_keys=True)
        except TypeError:
            # controller contains info that can't be json serialized:
            controller_args = str(data)

        program = self.program.name
        return "%s(%s)(%s)" % (controller_args, program, self.mimetype)



from cStringIO import StringIO
from giotto.programs import GiottoProgram
from giotto.views import GiottoView, renders

class FileView(GiottoView):
    @renders('*/*')
    def html(self, result):
        return {'body': result, 'mimetype': ''}

    @renders('text/x-cmd')
    def cmd(self, result):
        return {'body': result.read(), 'mimetype': ''}

def StaticServe(base_path):
    """
    Meta program for serving any file based on the path
    """
    def get_file(path):
        return open(base_path + path, 'r')

    class StaticServe(GiottoProgram):
        controllers = ('http-get', )
        model = [get_file, StringIO("Mock file content")]
        view = FileView

    return StaticServe()

def SingleStaticServe(file_path):
    """
    Meta program for serving a single file. Useful for favicon.ico
    """
    def get_file():
        return open(file_path, 'r')

    class SingleStaticServe(GiottoProgram):
        controllers = ('http-get', )
        model = [get_file, StringIO("Mock file content")]
        view = FileView

    return SingleStaticServe()


import sys 
import socket 
import string 
import os
import traceback

try:
  import irc.bot
except ImportError:
  raise ImportError("Requires irclib; pip install irc")

from giotto.controllers import GiottoController
from giotto.exceptions import ProgramNotFound
from giotto.utils import parse_kwargs

irc_execution_snippet = """
parser = argparse.ArgumentParser(description='Giotto Project Creator')
parser.add_argument('--model-mock', action='store_true', help='Mock out the model')
args = parser.parse_args()

config = {
    'host': '',
    'port': 6667,
    'nick': '',
    'ident': 'giotto',
    'realname': 'Giotto IRC Bot',
    'owner': '',
    'channels': '', # comma seperated
    'magic_token': '!giotto ',
}
from giotto.controllers.irc_ import listen
listen(manifest, config, model_mock=args.model_mock)"""

class IRCController(GiottoController):
    name = 'irc'
    default_mimetype = 'text/x-irc'

    def get_invocation(self):
        return self.request.program

    def get_controller_name(self):
        return 'irc'

    def get_raw_data(self):
        kwargs = self.request.args
        return parse_kwargs(kwargs)

    def get_concrete_response(self):
        try:
            result = self.get_data_response()
        except ProgramNotFound:
            result = {'body': "Program not found"}

        # convert to a format appropriate to the IRC Response api.
        return dict(
            response=result['body'],
            say_to=self.request.sent_to,
        )

    def get_primitive(self, primitive):
        if primitive == 'RAW_PAYLOAD':
            return self.get_data()


class IRCRequest(object):

    # the program name requested
    program = ''
    
    # the usr/channel the message was sent to
    sent_to = ''

    # PRIVMSG or whatever else...
    msg_type = ''

    # The message after the magic token has been removed
    # eg: !giotto multiply --x=1 --y=2  --> multiply --x=1 --y=2
    # note, invocations given through private message have no magic token
    # so this value will be the same as `message`
    message_token_removed = ''

    # the raw message with the magic token still attached
    raw_message = ''

    # boolean, was this request sent to a channel (True) or through private msg?
    channel_msg = None

    # opposite of `channel_msg`
    private_msg = None

    # the username of the person who made the request
    username = ''

    # the ident of the user who made the request
    ident = ''

    def __init__(self, event, magic_token, nick):
        self.ident = event.source()
        self.username = self.ident.split("!")[0]
        self.msg_type = event.eventtype()
        self.sent_to = event.target()
        self.private_message = self.msg_type == "privmsg"
        self.raw_message = event.arguments()[0]
        self.program, self.args = self.get_program_and_args(self.raw_message,magic_token)
        #print self.__repr__()

    def get_program_and_args(self, message, magic_token):
        if self.private_message == True:
            program = message.split()[0]
            args = message.split()[1:]
        else:
            # channel invocationa
            l = len(magic_token)
            parsed_message = message[l:]
            args = parsed_message.split()[1:]
            program = parsed_message.split()[0]
      
        return program, args

    def __repr__(self):
        return "program: %s, args: %s" % (self.program, self.args)

class IrcBot(irc.bot.SingleServerIRCBot):
    def __init__(self, config):
        if not config['host']:
            raise SystemExit('Error: IRC controller needs to be configured with a hostname')

        if not config['nick']:
            raise SystemExit('Error: IRC controller needs to be configured with a nick')

        print "Connecting to %s:%s as %s" % (config['host'],config['port'], config['nick'])

        irc.bot.SingleServerIRCBot.__init__(
            self, 
            [(config['host'],config['port'])], 
            config['nick'], 
            config['realname']
        )

        channels = config['channels']
        if channels:
            self.channel = channels
            print "Joining Channels: %s" % channels
        self.config = config
  
    def on_nicknameinuse(self, c, e):
        c.nick(c.get_nickname() + "_")

    def on_welcome(self, c, e):
        c.join(self.channel)

    def on_privmsg(self, c, e):
        self.process_message(c, e)

    def on_pubmsg(self, c, e):
        if e.arguments()[0].startswith(self.config['magic_token']) == False: return
        self.process_message(c, e)
 
    def process_message(self, c, e):
        request = IRCRequest(e,self.config['magic_token'],c.get_nickname())
        
        try:
            controller = IRCController(request, self.config['manifest'], self.config['model_mock'])
            result = controller.get_response()
        except Exception as exc:
            cls = exc.__class__.__name__
            c.privmsg(request.sent_to, "\x0304%s - %s: %s" % (request.program, cls, exc))
            traceback.print_exc(file=sys.stdout)
        else:
            msg = "%s: %s" % (request.username, result['response'])
            if request.private_message:
                c.privmsg(request.username, msg)
            else:
                c.privmsg(request.sent_to, msg)


def listen(manifest, config, model_mock=False):
    """
    IRC listening process.
    """
    config['manifest'] = manifest
    config['model_mock'] = model_mock
    IRC = IrcBot(config)
    IRC.start()



import re
import bcrypt

from giotto.exceptions import InvalidInput
from giotto.primitives import LOGGED_IN_USER
from giotto import config

from sqlalchemy import Column, String

class User(config.Base):
    username = Column(String, primary_key=True)
    password = Column(String)

    def __init__(self, username, password):
        self.username = username
        hashed = bcrypt.hashpw(password, bcrypt.gensalt())
        self.password = hashed
        self.raw_password = password
    
    def validate(self):
        """
        Make sure this newly created user instance meets the username/password
        requirements
        """
        r = getattr(config, 'auth_regex', r'^[\d\w]{4,30}$')
        errors = {}
        if not re.match(r, self.username):
            errors['username'] = {'message': 'Username not valid', 'value': self.username}
        if len(self.raw_password) < 4:
            errors['password'] = {'message': 'Password much be at least 4 characters'}
        if config.session.query(User).filter_by(username=self.username).first():
            errors['username'] = {'message': 'Username already exists', 'value': self.username}

        if errors:
            raise InvalidInput(data=errors)

    @classmethod
    def get_user_by_password(cls, username, password):
        """
        Given a username and a raw, unhashed password, get the corresponding
        user, retuns None if no match is found.
        """
        user = config.session.query(cls).filter_by(username=username).first()

        if not user:
            return None

        if bcrypt.hashpw(password, user.password) == user.password:
            return user
        else:
            return None

    @classmethod
    def get_user_by_hash(cls, username, hash_):
        return config.session.query(cls)\
                     .filter_by(username=username, password=hash_)\
                     .first()

    @classmethod
    def create(cls, username, password):
        """
        Create a new user instance
        """
        user = cls(username=username, password=password)
        user.validate()
        config.session.add(user)
        config.session.commit()
        return user

    @classmethod
    def all(cls):
        return config.session.query(cls).all()

    def __repr__(self):
        return "<User('%s', '%s')>" % (self.username, self.password)

def is_authenticated(message):
    """
    Is this request coming from a user who is logged in? This is a meta function.
    When adding this to a program, call it with the mesage you want to be displayed
    when no user is found.
    """
    def inner(user=LOGGED_IN_USER):
        if not user:
            raise InvalidInput(message)
        return {'user', user}
    return inner

def basic_register(username, password, password2):
    if password != password2:
        raise InvalidInput(data={'password': {'message': "Passwords do not match"},
                                 'username': {'value': username}})
    return User.create(username, password)





from collections import defaultdict
import datetime
import pickle
from sqlalchemy import Column, String, DateTime

try:
    import pylibmc
except ImportError:
    pylibmc = None

try:
    import redis
except ImportError:
    redis = None

DBKeyValue = None

class GiottoKeyValue(object):
    def set(self, key, obj):
        raise NotImplementedError

    def get(self, key, obj):
        raise NotImplementedError

class DatabaseKeyValue(GiottoKeyValue):
    def __init__(self, base, session):
        """
        Since we can't get the base class by importing it (base class defined in
        the config file, which is where this class is initialized). The base
        class must be passed in via this constructor. The SQLAlchemy model
        is then passed back into the mosule scope where it can be used
        by the DatabaseKeyValue backend.
        """
        if not base:
            return
        class _DBKeyValue(base):
            __tablename__ = 'giotto_keyvalue'
            key = Column(String, primary_key=True)
            value = Column(String)
            expires = Column(DateTime)

            @classmethod
            def set(cls, key, obj, expire):
                when_expire = datetime.datetime.now() + datetime.timedelta(seconds=expire)
                new = cls(key=key, value=pickle.dumps(obj), expires=when_expire)
                session.merge(new)
                session.commit()

            @classmethod
            def get(cls, key):
                value = session.query(cls)\
                               .filter_by(key=key)\
                               .filter(cls.expires > datetime.datetime.now())\
                               .first()
                
                if value:
                    return pickle.loads(str(value.value))
                return None

        global DBKeyValue
        DBKeyValue = _DBKeyValue

    def get(self, key):
        return DBKeyValue.get(key)

    def set(self, key, obj, expire):
        return DBKeyValue.set(key, obj, expire)

locmem = {}
class LocMemKeyValue(GiottoKeyValue):
    """
    KeyValue backend that stores everything in a python dict.
    """
    def get(self, key):
        if key not in locmem:
            return None

        result = locmem[key]
        expire = result[1]
        obj = result[0]

        if datetime.datetime.now() < expire:
            return obj

        del locmem[key]
        return None # obj has expired.

    def set(self, key, obj, expire):
        when_expire = datetime.datetime.now() + datetime.timedelta(seconds=expire)
        locmem[key] = (obj, when_expire)


class MemcacheKeyValue(GiottoKeyValue):
    def __init__(self, host=['localhost'], behavior={}):
        if not pylibmc:
            raise ImportError('pylibmc not installed! install with: pip install pylibmc')
        if hasattr(host, 'lower'):
            host = [host]
        kwargs = {'servers': host, 'binary': True}
        if behavior:
            kwargs['behavior'] = behavior
        self.client = pylibmc.Client(**kwargs)

    def set(self, key, obj, expire):
        self.client.set(key, obj, time=expire)

    def get(self, key):
        return self.client.get(key)

class RedisKeyValue(GiottoKeyValue):
    def __init__(self, host='localhost', port=6379, db=0):
        if not redis:
            raise ImportError('redis python wrapper not installed! install with: pip install redis')
        self.redis = redis.StrictRedis(host=host, port=port, db=db)

    def set(self, key, obj, expire):
        self.redis.setex(key, expire, pickle.dumps(obj))

    def get(self, key):
        pickled_value = self.redis.get(key)
        if pickled_value is None:
            return None
        return pickle.loads(pickled_value)

class DummyKeyValue(GiottoKeyValue):
    """
    Cache that does not save nor return a hit ever. Used as a placeholder.
    """
    def set(self, key, obj, expire):
        return None

    def get(self, key):
        return None


from giotto.programs import GiottoProgram
from giotto.views import BasicView
from giotto.primitives import LOGGED_IN_USER
from giotto.models import is_authenticated

from models import User
from views import BasicRegisterForm
from middleware import AuthenticationMiddleware, SetAuthenticationCookie, LogoutMiddleware

class UserRegistrationSubmit(GiottoProgram):
    name = 'register'
    controllers = ('http-post', 'cmd')
    model = [User.new_user, User(username='test', password='pass')]
    view = BasicRegisterForm

class UserRegistrationForm(GiottoProgram):
    name = 'register'
    controllers = ('http-get', )
    model = []
    view = BasicRegisterForm

class LoginForm(GiottoProgram):
    name = 'login'
    controllers = ('http-get', )
    model = []
    view = BasicRegisterForm

class LoginSubmit(GiottoProgram):
    """
    When the user submits the login form, the data get sent to this program.
    If their username/password match, the cookie is made and they are logged in.
    """
    name = 'login'
    controllers = ('http-post', )
    input_middleware = [AuthenticationMiddleware]
    model = [is_authenticated('Login credentials incorrect'), {'user': User('test', password='pass')}]
    view = BasicView
    output_middleware = [SetAuthenticationCookie]

class Logout(GiottoProgram):
    name = "logout"
    controllers = ('http-post', 'cmd')
    output_middleware = [LogoutMiddleware]

class ListAllUsers(GiottoProgram):
    name = 'all_users'
    controllers = ('http-get', 'cmd')
    model = [User.all]
    view = BasicView

manifest = ProgramManifest({
    'login': LoginSubmit(),
    'login_form': LoginForm(),
    'register': UserRegistrationSubmit(),
    'register_form': UserRegistrationForm(),
})

management_manifest = ProgramManifest({
    'list_all': ListAllUsers(),
})


from giotto.programs import GiottoProgram
import os

def shell(shell):
    from giotto import config
    
    if shell == "bpython":
        try:
            import bpython
            bpython.embed()
            return None
        except ImportError:
            pass

    elif shell == 'ipython':
        try:
            from IPython import embed
            embed()
            return None
        except ImportError:
            pass

    else:
        import code
        imported_objects = {'config': config}
        try:  # Try activating rlcompleter, because it's handy.
            import readline
        except ImportError:
            pass
        else:
            # We don't have to wrap the following import in a 'try', because
            # we already know 'readline' was imported successfully.
            import rlcompleter
            readline.set_completer(rlcompleter.Completer(imported_objects).complete)
            readline.parse_and_bind("tab:complete")

        # We want to honor both $PYTHONSTARTUP and .pythonrc.py, so follow system
        # conventions and get $PYTHONSTARTUP first then .pythonrc.py.
        for pythonrc in (os.environ.get("PYTHONSTARTUP"),
                         os.path.expanduser('~/.pythonrc.py')):
            if pythonrc and os.path.isfile(pythonrc):
                try:
                    with open(pythonrc) as handle:
                        exec(compile(handle.read(), pythonrc, 'exec'))
                except NameError:
                    pass
        code.interact(local=imported_objects)

    return {'body': "bpython or IPython required"}

class Shell(GiottoProgram):
    """
    Display a list of all instaled programs for all controllers for the
    currently invoked application.
    """
    controllers = ('cmd', )
    model = [shell]